@@ -1,3 +1,4 @@
+{-# LANGUAGE CPP #-}
 -----------------------------------------------------------------------------
 -- |
 -- Module      :  Distribution.Client.Config
@@ -33,7 +34,9 @@ module Distribution.Client.Config (
     haddockFlagsFields,
     installDirsFields,
     withProgramsFields,
-    withProgramOptionsFields
+    withProgramOptionsFields,
+    userConfigDiff,
+    userConfigUpdate
   ) where
 
 import Distribution.Client.Types
@@ -47,9 +50,11 @@ import Distribution.Client.Setup
          , UploadFlags(..), uploadCommand
          , ReportFlags(..), reportCommand
          , showRepo, parseRepo )
+import Distribution.Utils.NubList
+         ( NubList, fromNubList, toNubList)
 
 import Distribution.Simple.Compiler
-         ( OptimisationLevel(..) )
+         ( DebugInfoLevel(..), OptimisationLevel(..) )
 import Distribution.Simple.Setup
          ( ConfigFlags(..), configureOptions, defaultConfigFlags
          , HaddockFlags(..), haddockOptions, defaultHaddockFlags
@@ -77,20 +82,22 @@ import Distribution.Simple.Command
 import Distribution.Simple.Program
          ( defaultProgramConfiguration )
 import Distribution.Simple.Utils
-         ( notice, warn, lowercase )
+         ( die, notice, warn, lowercase, cabalVersion )
 import Distribution.Compiler
          ( CompilerFlavor(..), defaultCompilerFlavor )
 import Distribution.Verbosity
          ( Verbosity, normal )
 
 import Data.List
-         ( partition, find )
+         ( partition, find, foldl' )
 import Data.Maybe
          ( fromMaybe )
+#if !MIN_VERSION_base(4,8,0)
 import Data.Monoid
          ( Monoid(..) )
+#endif
 import Control.Monad
-         ( unless, foldM, liftM )
+         ( unless, foldM, liftM, liftM2 )
 import qualified Distribution.Compat.ReadP as Parse
          ( option )
 import qualified Text.PrettyPrint as Disp
@@ -109,6 +116,13 @@ import Distribution.Compat.Environment
          ( getEnvironment )
 import Distribution.Compat.Exception
          ( catchIO )
+import qualified Paths_cabal_install
+         ( version )
+import Data.Version
+         ( showVersion )
+import Data.Char
+         ( isSpace )
+import qualified Data.Map as M
 
 --
 -- * Configuration saved in the config file
@@ -178,13 +192,13 @@ instance Monoid SavedConfig where
         in case b' of [] -> a'
                       _  -> b'
 
-      lastNonEmptyNL' :: (SavedConfig -> flags) -> (flags -> [a])
-                      -> [a]
+      lastNonEmptyNL' :: (SavedConfig -> flags) -> (flags -> NubList a)
+                      -> NubList a
       lastNonEmptyNL' field subfield =
         let a' = subfield . field $ a
             b' = subfield . field $ b
-        in case b' of [] -> a'
-                      _  -> b'
+        in case fromNubList b' of [] -> a'
+                                  _  -> b'
 
       combinedSavedGlobalFlags = GlobalFlags {
         globalVersion           = combine globalVersion,
@@ -222,6 +236,7 @@ instance Monoid SavedConfig where
         installSummaryFile           = lastNonEmptyNL installSummaryFile,
         installLogFile               = combine installLogFile,
         installBuildReports          = combine installBuildReports,
+        installReportPlanningFailure = combine installReportPlanningFailure,
         installSymlinkBinDir         = combine installSymlinkBinDir,
         installOneShot               = combine installOneShot,
         installNumJobs               = combine installNumJobs,
@@ -249,6 +264,7 @@ instance Monoid SavedConfig where
         -- TODO: NubListify
         configConfigureArgs       = lastNonEmpty configConfigureArgs,
         configOptimization        = combine configOptimization,
+        configDebugInfo           = combine configDebugInfo,
         configProgPrefix          = combine configProgPrefix,
         configProgSuffix          = combine configProgSuffix,
         -- Parametrised by (Flag PathTemplate), so safe to use 'mappend'.
@@ -273,12 +289,16 @@ instance Monoid SavedConfig where
         configConstraints         = lastNonEmpty configConstraints,
         -- TODO: NubListify
         configDependencies        = lastNonEmpty configDependencies,
+        configInstantiateWith     = lastNonEmpty configInstantiateWith,
         -- TODO: NubListify
         configConfigurationsFlags = lastNonEmpty configConfigurationsFlags,
         configTests               = combine configTests,
         configBenchmarks          = combine configBenchmarks,
+        configCoverage            = combine configCoverage,
         configLibCoverage         = combine configLibCoverage,
-        configExactConfiguration  = combine configExactConfiguration
+        configExactConfiguration  = combine configExactConfiguration,
+        configFlagError           = combine configFlagError,
+        configRelocatable         = combine configRelocatable
         }
         where
           combine        = combine'        savedConfigureFlags
@@ -410,14 +430,14 @@ initialSavedConfig = do
   return mempty {
     savedGlobalFlags     = mempty {
       globalCacheDir     = toFlag cacheDir,
-      globalRemoteRepos  = [defaultRemoteRepo],
+      globalRemoteRepos  = toNubList [defaultRemoteRepo],
       globalWorldFile    = toFlag worldFile
     },
     savedConfigureFlags  = mempty {
-      configProgramPathExtra = extraPath
+      configProgramPathExtra = toNubList extraPath
     },
     savedInstallFlags    = mempty {
-      installSummaryFile = [toPathTemplate (logsDir </> "build.log")],
+      installSummaryFile = toNubList [toPathTemplate (logsDir </> "build.log")],
       installBuildReports= toFlag AnonymousReports,
       installNumJobs     = toFlag Nothing
     }
@@ -500,11 +520,9 @@ loadConfig verbosity configFileFlag userInstallFlag = addBaseConf $ do
       return conf
     Just (ParseFailed err) -> do
       let (line, msg) = locatedErrorMsg err
-      warn verbosity $
+      die $
           "Error parsing config file " ++ configFile
         ++ maybe "" (\n -> ':' : show n) line ++ ":\n" ++ msg
-      warn verbosity "Using default configuration."
-      initialSavedConfig
 
   where
     addBaseConf body = do
@@ -538,6 +556,9 @@ writeConfigFile file comments vals = do
       ,"-- Lines (like this one) beginning with '--' are comments."
       ,"-- Be careful with spaces and indentation because they are"
       ,"-- used to indicate layout for nested sections."
+      ,""
+      ,"-- Cabal library version: " ++ showVersion cabalVersion
+      ,"-- cabal-install version: " ++ showVersion Paths_cabal_install.version
       ,"",""
       ]
 
@@ -570,12 +591,12 @@ configFieldDescriptions :: [FieldDescr SavedConfig]
 configFieldDescriptions =
 
      toSavedConfig liftGlobalFlag
-       (commandOptions globalCommand ParseArgs)
+       (commandOptions (globalCommand []) ParseArgs)
        ["version", "numeric-version", "config-file", "sandbox-config-file"] []
 
   ++ toSavedConfig liftConfigFlag
        (configureOptions ParseArgs)
-       (["builddir", "configure-option", "constraint", "dependency"]
+       (["builddir", "constraint", "dependency"]
         ++ map fieldName installDirsFields)
 
         --FIXME: this is only here because viewAsFieldDescr gives us a parser
@@ -584,10 +605,11 @@ configFieldDescriptions =
        [simpleField "compiler"
           (fromFlagOrDefault Disp.empty . fmap Text.disp) (optional Text.parse)
           configHcFlavor (\v flags -> flags { configHcFlavor = v })
-        -- TODO: The following is a temporary fix. The "optimization" field is
-        -- OptArg, and viewAsFieldDescr fails on that. Instead of a hand-written
-        -- hackaged parser and printer, we should handle this case properly in
-        -- the library.
+        -- TODO: The following is a temporary fix. The "optimization"
+        -- and "debug-info" fields are OptArg, and viewAsFieldDescr
+        -- fails on that. Instead of a hand-written hackaged parser
+        -- and printer, we should handle this case properly in the
+        -- library.
        ,liftField configOptimization (\v flags -> flags { configOptimization = v }) $
         let name = "optimization" in
         FieldDescr name
@@ -609,6 +631,29 @@ configFieldDescriptions =
                lstr = lowercase str
                caseWarning = PWarning $
                  "The '" ++ name ++ "' field is case sensitive, use 'True' or 'False'.")
+       ,liftField configDebugInfo (\v flags -> flags { configDebugInfo = v }) $
+        let name = "debug-info" in
+        FieldDescr name
+          (\f -> case f of
+                   Flag NoDebugInfo      -> Disp.text "False"
+                   Flag MinimalDebugInfo -> Disp.text "1"
+                   Flag NormalDebugInfo  -> Disp.text "True"
+                   Flag MaximalDebugInfo -> Disp.text "3"
+                   _                     -> Disp.empty)
+          (\line str _ -> case () of
+           _ |  str == "False" -> ParseOk [] (Flag NoDebugInfo)
+             |  str == "True"  -> ParseOk [] (Flag NormalDebugInfo)
+             |  str == "0"     -> ParseOk [] (Flag NoDebugInfo)
+             |  str == "1"     -> ParseOk [] (Flag MinimalDebugInfo)
+             |  str == "2"     -> ParseOk [] (Flag NormalDebugInfo)
+             |  str == "3"     -> ParseOk [] (Flag MaximalDebugInfo)
+             | lstr == "false" -> ParseOk [caseWarning] (Flag NoDebugInfo)
+             | lstr == "true"  -> ParseOk [caseWarning] (Flag NormalDebugInfo)
+             | otherwise       -> ParseFailed (NoParse name line)
+             where
+               lstr = lowercase str
+               caseWarning = PWarning $
+                 "The '" ++ name ++ "' field is case sensitive, use 'True' or 'False'.")
        ]
 
   ++ toSavedConfig liftConfigExFlag
@@ -648,7 +693,8 @@ deprecatedFieldDescriptions =
   [ liftGlobalFlag $
     listField "repos"
       (Disp.text . showRepo) parseRepo
-      globalRemoteRepos (\rs cfg -> cfg { globalRemoteRepos = rs })
+      (fromNubList . globalRemoteRepos)
+      (\rs cfg -> cfg { globalRemoteRepos = toNubList rs })
   , liftGlobalFlag $
     simpleField "cachedir"
       (Disp.text . fromFlagOrDefault "") (optional parseFilePathQ)
@@ -721,10 +767,7 @@ parseConfig initial = \str -> do
        configProgramPaths  = paths,
        configProgramArgs   = args
        },
-    savedHaddockFlags      = haddockFlags {
-       haddockProgramPaths = paths,
-       haddockProgramArgs  = args
-       },
+    savedHaddockFlags      = haddockFlags,
     savedUserInstallDirs   = user,
     savedGlobalInstallDirs = global
   }
@@ -829,3 +872,60 @@ withProgramOptionsFields :: [FieldDescr [(String, [String])]]
 withProgramOptionsFields =
   map viewAsFieldDescr $
   programConfigurationOptions defaultProgramConfiguration ParseArgs id (++)
+
+-- | Get the differences (as a pseudo code diff) between the user's
+-- '~/.cabal/config' and the one that cabal would generate if it didn't exist.
+userConfigDiff :: GlobalFlags -> IO [String]
+userConfigDiff globalFlags = do
+  userConfig <- loadConfig normal (globalConfigFile globalFlags) mempty
+  testConfig <- liftM2 mappend baseSavedConfig initialSavedConfig
+  return $ reverse . foldl' createDiff [] . M.toList
+                $ M.unionWith combine
+                    (M.fromList . map justFst $ filterShow testConfig)
+                    (M.fromList . map justSnd $ filterShow userConfig)
+  where
+    justFst (a, b) = (a, (Just b, Nothing))
+    justSnd (a, b) = (a, (Nothing, Just b))
+
+    combine (Nothing, Just b) (Just a, Nothing) = (Just a, Just b)
+    combine (Just a, Nothing) (Nothing, Just b) = (Just a, Just b)
+    combine x y = error $ "Can't happen : userConfigDiff " ++ show x ++ " " ++ show y
+
+    createDiff :: [String] -> (String, (Maybe String, Maybe String)) -> [String]
+    createDiff acc (key, (Just a, Just b))
+        | a == b = acc
+        | otherwise = ("+ " ++ key ++ ": " ++ b) : ("- " ++ key ++ ": " ++ a) : acc
+    createDiff acc (key, (Nothing, Just b)) = ("+ " ++ key ++ ": " ++ b) : acc
+    createDiff acc (key, (Just a, Nothing)) = ("- " ++ key ++ ": " ++ a) : acc
+    createDiff acc (_, (Nothing, Nothing)) = acc
+
+    filterShow :: SavedConfig -> [(String, String)]
+    filterShow cfg = map keyValueSplit
+        . filter (\s -> not (null s) && any (== ':') s)
+        . map nonComment
+        . lines
+        $ showConfig cfg
+
+    nonComment [] = []
+    nonComment ('-':'-':_) = []
+    nonComment (x:xs) = x : nonComment xs
+
+    topAndTail = reverse . dropWhile isSpace . reverse . dropWhile isSpace
+
+    keyValueSplit s =
+        let (left, right) = break (== ':') s
+        in (topAndTail left, topAndTail (drop 1 right))
+
+
+-- | Update the user's ~/.cabal/config' keeping the user's customizations.
+userConfigUpdate :: Verbosity -> GlobalFlags -> IO ()
+userConfigUpdate verbosity globalFlags = do
+  userConfig <- loadConfig normal (globalConfigFile globalFlags) mempty
+  newConfig <- liftM2 mappend baseSavedConfig initialSavedConfig
+  commentConf <- commentSavedConfig
+  cabalFile <- defaultConfigFile
+  let backup = cabalFile ++ ".backup"
+  notice verbosity $ "Renaming " ++ cabalFile ++ " to " ++ backup ++ "."
+  renameFile cabalFile backup
+  notice verbosity $ "Writing merged config to " ++ cabalFile ++ "."
+  writeConfigFile cabalFile commentConf (newConfig `mappend` userConfig)
